# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Data Format
    ~~~~~~~~~~~

    Base64, Base58, Hex, UTF-8, JsON
"""

from abc import ABC, abstractmethod
from typing import Optional, Union, Any, Dict, List


class StringCoder(ABC):
    """
        String Coder
        ~~~~~~~~~~~~
        UTF-8, UTF-16, GBK, GB2312, ...

        1. encode string to binary data;
        2. decode binary data to string.
    """

    @abstractmethod
    def encode(self, string: str) -> bytes:
        """
        Encode local string to binary data

        :param string: local string
        :return: binary data
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, data: bytes) -> Optional[str]:
        """
        Decode binary data to local string

        :param data: binary data
        :return: local string
        """
        raise NotImplemented


class UTF8:
    coder: StringCoder = None

    @staticmethod
    def encode(string: str) -> bytes:
        # assert UTF8.coder is not None, 'UTF8 parser not set yet'
        return UTF8.coder.encode(string=string)

    @staticmethod
    def decode(data: bytes) -> Optional[str]:
        # assert UTF8.coder is not None, 'UTF8 parser not set yet'
        return UTF8.coder.decode(data=data)


class ObjectCoder(ABC):
    """
        Object Coder
        ~~~~~~~~~~~~
        JsON, XML, ...

        1. encode object to string;
        2. decode string to object.
    """

    @abstractmethod
    def encode(self, obj: Any) -> str:
        """
        Encode Map/List object to str

        :param obj: Map or List
        :return: serialized string
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, string: str) -> Optional[Any]:
        """
        Decode str to Map/List object

        :param string: serialized string
        :return: Map or List
        """
        raise NotImplemented


class JSON:
    coder: ObjectCoder = None

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSON.coder is not None, 'JSON parser not set yet'
        return JSON.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSON.coder is not None, 'JSON parser not set yet'
        return JSON.coder.decode(string=string)


class MapCoder(ObjectCoder, ABC):
    """ coder for json <=> map """

    # Override
    def encode(self, obj: Dict) -> str:
        return JSON.encode(obj=obj)

    # Override
    def decode(self, string: str) -> Optional[Dict]:
        return JSON.decode(string=string)


class ListCoder(ObjectCoder, ABC):
    """ coder for json <=> list """

    # Override
    def encode(self, obj: List) -> str:
        return JSON.encode(obj=obj)

    # Override
    def decode(self, string: str) -> Optional[List]:
        return JSON.decode(string=string)


class JSONMap:
    coder = MapCoder()

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSONMap.coder is not None, 'JSONMap parser not set yet'
        return JSONMap.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSONMap.coder is not None, 'JSONMap parser not set yet'
        return JSONMap.coder.decode(string=string)


class JSONList:
    coder = ListCoder()

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSONList.coder is not None, 'JSONList parser not set yet'
        return JSONList.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSONList.coder is not None, 'JSONList parser not set yet'
        return JSONList.coder.decode(string=string)


class DataCoder(ABC):
    """
        Data Coder
        ~~~~~~~~~~
        Hex, Base58, Base64, ...

        1. encode binary data to string;
        2. decode string to binary data.
    """

    @abstractmethod
    def encode(self, data: bytes) -> str:
        """
        Encode binary data to text string

        :param data: binary data
        :return:     text string (Base58/64, Hex, ...)
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, string: str) -> Optional[bytes]:
        """
        Decode text string to binary data

        :param string: text string (Base58/64, Hex, ...)
        :return:       binary data
        """
        raise NotImplemented


class Hex:
    coder: DataCoder = None

    @staticmethod
    def encode(data: bytes) -> str:
        # assert Hex.coder is not None, 'Hex coder not set yet'
        return Hex.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        # assert Hex.coder is not None, 'Hex coder not set yet'
        return Hex.coder.decode(string=string)


class Base58:
    coder: DataCoder = None

    @staticmethod
    def encode(data: bytes) -> str:
        # assert Base58.coder is not None, 'Base58 coder not set yet'
        return Base58.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        # assert Base58.coder is not None, 'Base58 coder not set yet'
        return Base58.coder.decode(string=string)


class Base64:
    coder: DataCoder = None

    @staticmethod
    def encode(data: bytes) -> str:
        # assert Base64.coder is not None, 'Base64 coder not set yet'
        return Base64.coder.encode(data=data)

    @staticmethod
    def decode(string: str) -> Optional[bytes]:
        # assert Base64.coder is not None, 'Base64 coder not set yet'
        return Base64.coder.decode(string=string)


#
#   Interfaces
#


def base64_encode(data: bytes) -> str:
    return Base64.encode(data)


def base64_decode(string: str) -> Optional[bytes]:
    return Base64.decode(string)


def base58_encode(data: bytes) -> str:
    return Base58.encode(data)


def base58_decode(string: str) -> Optional[bytes]:
    return Base58.decode(string)


def hex_encode(data: bytes) -> str:
    return Hex.encode(data)


def hex_decode(string: str) -> Optional[bytes]:
    return Hex.decode(string)


def json_encode(obj: Union[Dict, List]) -> str:
    return JSON.encode(obj=obj)


def json_decode(string: str) -> Union[Dict, List, None]:
    return JSON.decode(string=string)


def utf8_encode(string: str) -> bytes:
    return UTF8.encode(string=string)


def utf8_decode(data: bytes) -> Optional[str]:
    return UTF8.decode(data=data)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2022 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2022 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from enum import IntEnum


class EntityType(IntEnum):
    """
        @enum MKMEntityType

        @abstract A network ID to indicate what kind the entity is.

        @discussion An address can identify a person, a group of people,
            a team, even a thing.

            MKMEntityType_User indicates this entity is a person's account.
            An account should have a public key, which proved by meta data.

            MKMEntityType_Group indicates this entity is a group of people,
            which should have a founder (also the owner), and some members.

            MKMEntityType_Station indicates this entity is a DIM network station.

            MKMEntityType_ISP indicates this entity is a group for stations.

            MKMEntityType_Bot indicates this entity is a bot user.

            MKMEntityType_Company indicates a company for stations and/or bots.

        Bits:
            0000 0001 - group flag
            0000 0010 - node flag
            0000 0100 - bot flag
            0000 1000 - CA flag
            ...         (reserved)
            0100 0000 - customized flag
            1000 0000 - broadcast flag

            (All above are just some advices to help choosing numbers :P)
    """
    ################################
    #  Main: 0, 1
    ################################
    USER = 0x00             # 0000 0000
    GROUP = 0x01            # 0000 0001 (User Group)

    ################################
    #  Network: 2, 3
    ################################
    STATION = 0x02          # 0000 0010 (Server Node)
    ISP = 0x03              # 0000 0011 (Service Provider)
    # STATION_GROUP = 0x03  # 0000 0011

    ################################
    #  Bot: 4, 5
    ################################
    BOT = 0x04              # 0000 0100 (Business Node)
    ICP = 0x05              # 0000 0101 (Content Provider)
    # BOT_GROUP = 0x05      # 0000 0101

    ################################
    #  Management: 6, 7, 8
    ################################
    SUPERVISOR = 0x06       # 0000 0110 (Company President)
    COMPANY = 0x07          # 0000 0111 (Super Group for ISP/ICP)
    # CA = 0x08             # 0000 1000 (Certification Authority)

    ################################
    #  Customized: 64, 65
    ################################
    # APP_USER = 0x40       # 0100 0000 (Application Customized User)
    # APP_GROUP = 0x41      # 0100 0001 (Application Customized Group)

    ################################
    #  Broadcast: 128, 129
    ################################
    ANY = 0x80              # 1000 0000 (anyone@anywhere)
    EVERY = 0x81            # 1000 0001 (everyone@everywhere)

    @classmethod
    def is_user(cls, network: int) -> bool:
        return (network & cls.GROUP) == cls.USER

    @classmethod
    def is_group(cls, network: int) -> bool:
        return (network & cls.GROUP) == cls.GROUP

    @classmethod
    def is_broadcast(cls, network: int) -> bool:
        return (network & cls.ANY) == cls.ANY



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Any, Dict

from ..types import Wrapper

from .cryptography import EncryptKey, DecryptKey
from .symmetric import SymmetricKey, SymmetricKeyFactory
from .asymmetric import SignKey, VerifyKey
from .public import PublicKey, PublicKeyFactory
from .private import PrivateKey, PrivateKeyFactory


class CryptographyKeyGeneralFactory:

    # sample data for checking keys
    promise = 'Moky loves May Lee forever!'.encode('utf-8')

    def __init__(self):
        super().__init__()
        # str(algorithm) -> SymmetricKey.Factory
        self.__symmetric_key_factories: Dict[str, SymmetricKeyFactory] = {}
        # str(algorithm) -> PublicKey.Factory
        self.__public_key_factories: Dict[str, PublicKeyFactory] = {}
        # str(algorithm) -> PrivateKey.Factory
        self.__private_key_factories: Dict[str, PrivateKeyFactory] = {}

    def keys_match(self, encrypt_key: EncryptKey, decrypt_key: DecryptKey) -> bool:
        """ check by encryption """
        ciphertext = encrypt_key.encrypt(data=self.promise)
        plaintext = decrypt_key.decrypt(data=ciphertext)
        return plaintext == self.promise

    def asymmetric_keys_match(self, sign_key: SignKey, verify_key: VerifyKey) -> bool:
        """ verify with signature """
        signature = sign_key.sign(data=self.promise)
        return verify_key.verify(data=self.promise, signature=signature)

    # noinspection PyMethodMayBeStatic
    def get_key_algorithm(self, key: Dict[str, Any]) -> Optional[str]:
        """ get key algorithm name """
        return key.get('algorithm')

    #
    #   SymmetricKey
    #

    def set_symmetric_key_factory(self, algorithm: str, factory: SymmetricKeyFactory):
        self.__symmetric_key_factories[algorithm] = factory

    def get_symmetric_key_factory(self, algorithm: str) -> Optional[SymmetricKeyFactory]:
        return self.__symmetric_key_factories.get(algorithm)

    def generate_symmetric_key(self, algorithm: str) -> Optional[SymmetricKey]:
        factory = self.get_symmetric_key_factory(algorithm=algorithm)
        # assert factory is not None, 'key algorithm not support: %s' % algorithm
        return factory.generate_symmetric_key()

    def parse_symmetric_key(self, key: Any) -> Optional[SymmetricKey]:
        if key is None:
            return None
        if isinstance(key, SymmetricKey):
            return key
        info = Wrapper.get_dict(key)
        if info is None:
            # assert False, 'key error: %s' % key
            return None
        algorithm = self.get_key_algorithm(key=info)
        if algorithm is None:
            algorithm = '*'
        factory = self.get_symmetric_key_factory(algorithm=algorithm)
        if factory is None and algorithm != '*':
            factory = self.get_symmetric_key_factory(algorithm='*')  # unknown
        # if factory is None:
        #     # assert False, 'key algorithm not support: %s' % algorithm
        #     return None
        return factory.parse_symmetric_key(key=info)

    #
    #   PublicKey
    #

    def set_public_key_factory(self, algorithm: str, factory: PublicKeyFactory):
        self.__public_key_factories[algorithm] = factory

    def get_public_key_factory(self, algorithm: str) -> Optional[PublicKeyFactory]:
        return self.__public_key_factories.get(algorithm)

    def parse_public_key(self, key: Any) -> Optional[PublicKey]:
        if key is None:
            return None
        elif isinstance(key, PublicKey):
            return key
        info = Wrapper.get_dict(key)
        if info is None:
            # assert False, 'key error: %s' % key
            return None
        algorithm = self.get_key_algorithm(key=info)
        if algorithm is None:
            algorithm = '*'
        factory = self.get_public_key_factory(algorithm=algorithm)
        if factory is None and algorithm != '*':
            factory = self.get_public_key_factory(algorithm='*')  # unknown
        # if factory is None:
        #     # assert False, 'key algorithm not support: %s' % algorithm
        #     return None
        return factory.parse_public_key(key=info)

    #
    #   PrivateKey
    #

    def set_private_key_factory(self, algorithm: str, factory: PrivateKeyFactory):
        self.__private_key_factories[algorithm] = factory

    def get_private_key_factory(self, algorithm: str) -> Optional[PrivateKeyFactory]:
        return self.__private_key_factories.get(algorithm)

    def generate_private_key(self, algorithm: str) -> Optional[PrivateKey]:
        factory = self.get_private_key_factory(algorithm=algorithm)
        # assert factory is not None, 'key algorithm not support: %s' % algorithm
        return factory.generate_private_key()

    def parse_private_key(self, key: Any) -> Optional[PrivateKey]:
        if key is None:
            return None
        elif isinstance(key, PrivateKey):
            return key
        info = Wrapper.get_dict(key)
        if info is None:
            # assert False, 'key error: %s' % key
            return None
        algorithm = self.get_key_algorithm(key=info)
        if algorithm is None:
            algorithm = '*'
        factory = self.get_private_key_factory(algorithm=algorithm)
        if factory is None and algorithm != '*':
            factory = self.get_private_key_factory(algorithm='*')  # unknown
        # if factory is None:
        #     # assert False, 'key algorithm not support: %s' % algorithm
        #     return None
        return factory.parse_private_key(key=info)


# Singleton
class CryptographyKeyFactoryManager:

    general_factory = CryptographyKeyGeneralFactory()



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union, List

from ..crypto import EncryptKey, VerifyKey

from .identifier import ID
from .tai_doc import Document


class Visa(Document, ABC):
    """
        User Document
        ~~~~~~~~~~~~~
        This interface is defined for authorizing other apps to login,
        which can generate a temporary asymmetric key pair for messaging.
    """

    @property
    @abstractmethod
    def key(self) -> Union[EncryptKey, VerifyKey, None]:
        """
        Get public key to encrypt message for user

        :return: public key
        """
        raise NotImplemented

    @key.setter
    @abstractmethod
    def key(self, value: Union[EncryptKey, VerifyKey]):
        """
        Set public key for other user to encrypt message

        :param value: public key as visa.key
        """
        raise NotImplemented

    @property
    @abstractmethod
    def avatar(self) -> Optional[str]:
        """
        Get avatar URL

        :return: URL string
        """
        raise NotImplemented

    @avatar.setter
    @abstractmethod
    def avatar(self, url: str):
        """
        Set avatar URL

        :param url: URL string
        """
        raise NotImplemented


class Bulletin(Document, ABC):
    """
        Group Document
        ~~~~~~~~~~~~~~
    """

    @property
    @abstractmethod
    def founder(self) -> Optional[ID]:
        """
        Get group founder

        :return: user ID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def assistants(self) -> Optional[List[ID]]:
        """
        Get group assistants

        :return: bot ID list
        """
        raise NotImplemented

    @assistants.setter
    @abstractmethod
    def assistants(self, bots: List[ID]):
        """
        Set group assistants

        :param bots: bot ID list
        """
        raise NotImplemented

    """
        Times
        ~~~~~
    """

    @property
    @abstractmethod
    def created_time(self) -> Optional[float]:
        raise NotImplemented

    @property
    @abstractmethod
    def modified_time(self) -> Optional[float]:
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Union, Any, List, Dict

from .types import Wrapper
from .crypto import SignKey, VerifyKey
from .crypto import utf8_encode
from .protocol import Address, AddressFactory
from .protocol import ID, IDFactory
from .protocol import Meta, MetaType, MetaFactory
from .protocol import Document, DocumentFactory


class AccountGeneralFactory:

    def __init__(self):
        super().__init__()
        # AddressFactory
        self.__address_factory: Optional[AddressFactory] = None
        # IDFactory
        self.__id_factory: Optional[IDFactory] = None
        # int(type) -> MetaFactory
        self.__meta_factories: Dict[int, MetaFactory] = {}
        # str(type) -> DocumentFactory
        self.__document_factories: Dict[str, DocumentFactory] = {}

    #
    #   Address
    #

    def set_address_factory(self, factory: AddressFactory):
        self.__address_factory = factory

    def get_address_factory(self) -> Optional[AddressFactory]:
        return self.__address_factory

    def generate_address(self, meta: Meta, network: int) -> Address:
        factory = self.get_address_factory()
        # assert factory is not None, 'address factory not set'
        return factory.generate_address(meta=meta, network=network)

    def create_address(self, address: str) -> Optional[AddressFactory]:
        factory = self.get_address_factory()
        # assert factory is not None, 'address factory not set'
        return factory.create_address(address=address)

    def parse_address(self, address: Any) -> Optional[Address]:
        if address is None:
            return None
        elif isinstance(address, Address):
            return address
        string = Wrapper.get_str(address)
        if string is None:
            # assert False, 'address error: %s' % address
            return None
        factory = self.get_address_factory()
        # assert factory is not None, 'address factory not set'
        return factory.parse_address(address=string)

    #
    #   ID
    #

    def set_id_factory(self, factory: IDFactory):
        self.__id_factory = factory

    def get_id_factory(self) -> Optional[IDFactory]:
        return self.__id_factory

    def generate_id(self, meta: Meta, network: int, terminal: Optional[str]) -> ID:
        factory = self.get_id_factory()
        # assert factory is not None, 'ID factory not set'
        return factory.generate_id(meta=meta, network=network, terminal=terminal)

    def create_id(self, name: Optional[str], address: Address, terminal: Optional[str]) -> ID:
        factory = self.get_id_factory()
        # assert factory is not None, 'ID factory not set'
        return factory.create_id(name=name, address=address, terminal=terminal)

    def parse_id(self, identifier: Any) -> Optional[ID]:
        if identifier is None:
            return None
        elif isinstance(identifier, ID):
            return identifier
        string = Wrapper.get_str(identifier)
        if string is None:
            # assert False, 'ID error: %s' % identifier
            return None
        factory = self.get_id_factory()
        # assert factory is not None, 'ID factory not set'
        return factory.parse_id(identifier=string)

    def convert_id_list(self, array: List[str]) -> List[ID]:
        """
        Convert ID list from string array

        :param array: ID string array
        :return: ID list
        """
        result = []
        for item in array:
            identifier = self.parse_id(identifier=item)
            if identifier is None:
                # id error
                continue
            result.append(identifier)
        return result

    # noinspection PyMethodMayBeStatic
    def revert_id_list(self, array: List[ID]) -> List[str]:
        """
        Revert ID list to string array

        :param array: ID list
        :return: string array
        """
        result = []
        for item in array:
            result.append(str(item))
        return result

    #
    #   Meta
    #

    def set_meta_factory(self, version: Union[MetaType, int], factory):
        if isinstance(version, MetaType):
            version = version.value
        self.__meta_factories[version] = factory

    def get_meta_factory(self, version: Union[MetaType, int]) -> Optional[MetaFactory]:
        if isinstance(version, MetaType):
            version = version.value
        return self.__meta_factories.get(version)

    # noinspection PyMethodMayBeStatic
    def get_meta_type(self, meta: Dict[str, Any]) -> Optional[int]:
        """ get meta type(version) """
        return meta.get('type')

    def generate_meta(self, version: Union[MetaType, int], key: SignKey,
                      seed: Optional[str]) -> Meta:
        factory = self.get_meta_factory(version=version)
        # assert factory is not None, 'failed to get meta factory: %d' % version
        return factory.generate_meta(key=key, seed=seed)

    def create_meta(self, version: Union[MetaType, int], key: VerifyKey,
                    seed: Optional[str], fingerprint: Union[bytes, str, None]) -> Meta:
        factory = self.get_meta_factory(version=version)
        # assert factory is not None, 'failed to get meta factory: %d' % version
        return factory.create_meta(key=key, seed=seed, fingerprint=fingerprint)

    def parse_meta(self, meta: Any) -> Optional[Meta]:
        if meta is None:
            return None
        elif isinstance(meta, Meta):
            return meta
        info = Wrapper.get_dict(meta)
        if info is None:
            # assert False, 'meta error: %s' % meta
            return None
        version = self.get_meta_type(meta=info)
        if version is None:
            version = 0
        factory = self.get_meta_factory(version=version)
        if factory is None and version != 0:
            factory = self.get_meta_factory(version=0)  # unknown
        # if factory is None:
        #     # assert False, 'meta factory not found: %d' % version
        #     return None
        return factory.parse_meta(meta=info)

    # noinspection PyMethodMayBeStatic
    def check_meta(self, meta: Meta) -> bool:
        key = meta.key
        if not MetaType.has_seed(version=meta.type):
            # this meta has no seed, so no signature too
            return True
            # return isinstance(key, VerifyKey)
        # check seed with signature
        seed = meta.seed
        fingerprint = meta.fingerprint
        if seed is None or fingerprint is None:
            # seed and fingerprint should not be empty
            return False
        # verify fingerprint
        return key.verify(data=utf8_encode(string=seed), signature=fingerprint)

    # noinspection PyMethodMayBeStatic
    def meta_match_id(self, meta: Meta, identifier: ID) -> bool:
        """ Check whether meta match with entity ID
            (must call this when received a new meta from network) """
        # check ID.name
        if meta.seed == identifier.name:
            # check ID.address
            old = identifier.address
            gen = Address.generate(meta=meta, network=old.type)
            return old == gen

    # noinspection PyMethodMayBeStatic
    def meta_match_key(self, meta: Meta, key: VerifyKey) -> bool:
        """ Check whether meta match with public key """
        if key == meta.key:
            # NOTICE: ID with BTC/ETH address has no username, so
            #         just compare the key.data to check matching
            return True
        # check with seed & fingerprint
        if MetaType.has_seed(version=meta.type):
            # check whether keys equal by verifying signature
            seed = utf8_encode(string=meta.seed)
            fingerprint = meta.fingerprint
            return key.verify(data=seed, signature=fingerprint)

    #
    #   Document
    #

    def set_document_factory(self, doc_type: str, factory):
        self.__document_factories[doc_type] = factory

    def get_document_factory(self, doc_type: str) -> Optional[DocumentFactory]:
        return self.__document_factories.get(doc_type)

    # noinspection PyMethodMayBeStatic
    def get_document_type(self, document: Dict[str, Any]) -> Optional[str]:
        return document.get('type')

    def create_document(self, doc_type: str, identifier: ID,
                        data: Optional[str], signature: Union[bytes, str]) -> Document:
        factory = self.get_document_factory(doc_type=doc_type)
        # assert factory is not None, 'document factory not found for type: %s' % doc_type
        return factory.create_document(identifier=identifier, data=data, signature=signature)

    def parse_document(self, document: Any) -> Optional[Document]:
        if document is None:
            return None
        elif isinstance(document, Document):
            return document
        info = Wrapper.get_dict(document)
        if info is None:
            # assert False, 'document error: %s' % document
            return None
        doc_type = self.get_document_type(document=info)
        if doc_type is None:
            doc_type = '*'
        factory = self.get_document_factory(doc_type=doc_type)
        if factory is None and doc_type != '*':
            factory = self.get_document_factory(doc_type='*')  # unknown
        # if factory is None:
        #     # assert False, 'document factory not found for type: %s' % doc_type
        #     return None
        return factory.parse_document(document=info)


# Singleton
class AccountFactoryManager:

    general_factory = AccountGeneralFactory()



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Union

from .types import ConstantString
from .protocol import EntityType
from .protocol import Address


"""
    Address for Broadcast
    ~~~~~~~~~~~~~~~~~~~~~
"""


class BroadcastAddress(ConstantString, Address):

    def __init__(self, address: str, network: Union[int, EntityType]):
        super().__init__(string=address)
        if isinstance(network, EntityType):
            network = network.value
        self.__type = network

    @property  # Override
    def type(self) -> int:
        return self.__type

    @property  # Override
    def is_broadcast(self) -> bool:
        return True

    @property  # Override
    def is_user(self) -> bool:
        return self.type == EntityType.ANY

    @property  # Override
    def is_group(self) -> bool:
        return self.type == EntityType.EVERY


ANYWHERE = BroadcastAddress(address='anywhere', network=EntityType.ANY)
EVERYWHERE = BroadcastAddress(address='everywhere', network=EntityType.EVERY)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union, Any, Dict

from ..types import Mapper

from .identifier import ID
from .tai import TAI


class Document(TAI, Mapper, ABC):

    #
    #  Document types
    #
    VISA = 'visa'          # for login/communication
    PROFILE = 'profile'    # for user info
    BULLETIN = 'bulletin'  # for group info

    @property
    @abstractmethod
    def type(self) -> Optional[str]:
        """
        Get document type

        :return: doc type
        """
        raise NotImplemented

    @property
    @abstractmethod
    def identifier(self) -> ID:
        """
        Get entity ID

        :return: Entity ID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def time(self) -> float:
        """
        Get sign time

        :return: timestamp
        """
        raise NotImplemented

    #
    #  properties getter/setter
    #
    @property
    @abstractmethod
    def name(self) -> str:
        """
        Get entity name

        :return: name string
        """
        raise NotImplemented

    @name.setter
    @abstractmethod
    def name(self, string: str):
        """
        Set entity name

        :param string: name string
        :return:
        """
        raise NotImplemented

    #
    #   Factory Methods
    #

    @classmethod
    def create(cls, doc_type: str, identifier: ID,
               data: Optional[str] = None, signature: Union[bytes, str] = None):  # -> Optional[Document]:
        gf = general_factory()
        return gf.create_document(doc_type=doc_type, identifier=identifier, data=data, signature=signature)

    @classmethod
    def parse(cls, document: Any):  # -> Optional[Document]:
        gf = general_factory()
        return gf.parse_document(document=document)

    @classmethod
    def register(cls, doc_type: str, factory):
        gf = general_factory()
        gf.set_document_factory(doc_type=doc_type, factory=factory)

    @classmethod
    def factory(cls, doc_type: str):  # -> Optional[DocumentFactory]:
        gf = general_factory()
        return gf.get_document_factory(doc_type=doc_type)


def general_factory():
    from ..factory import AccountFactoryManager
    return AccountFactoryManager.general_factory


class DocumentFactory(ABC):

    @abstractmethod
    def create_document(self, identifier: ID, data: Optional[str], signature: Union[bytes, str, None]) -> Document:
        """
        1. Create a new empty document with entity ID

        2. Create document with data & signature loaded from local storage

        :param identifier: entity ID
        :param data:       document data
        :param signature:  document signature
        :return: Document
        """
        raise NotImplemented

    @abstractmethod
    def parse_document(self, document: Dict[str, Any]) -> Optional[Document]:
        """
        Parse map object to entity document

        :param document:
        :return:
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from enum import IntEnum


class MetaType(IntEnum):
    """
        @enum MKMMetaVersion

        @abstract Defined for meta data structure to generate identifier.

        @discussion Generate & check ID/Address

            MKMMetaVersion_MKM give a seed string first, and sign this seed to get
            fingerprint; after that, use the fingerprint to generate address.
            This will get a firmly relationship between (username, address & key).

            MKMMetaVersion_BTC use the key data to generate address directly.
            This can build a BTC address for the entity ID without username.

            MKMMetaVersion_ExBTC use the key data to generate address directly, and
            sign the seed to get fingerprint (just for binding username & key).
            This can build an entity ID with username and BTC address.

        Bits:
            0000 0001 - this meta contains seed as ID.name
            0000 0010 - this meta generate BTC address
            0000 0100 - this meta generate ETH address
            ...
    """
    DEFAULT = 0x01
    MKM = 0x01    # 0000 0001

    BTC = 0x02    # 0000 0010
    ExBTC = 0x03  # 0000 0011

    ETH = 0x04    # 0000 0100
    ExETH = 0x05  # 0000 0101

    @classmethod
    def has_seed(cls, version: int) -> bool:
        return (version & cls.MKM) == cls.MKM



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Any, Optional

from ..types import Stringer


class Address(Stringer, ABC):
    """This class is used to build address for ID

        Address for MKM ID
        ~~~~~~~~~~~~~~~~~~

        properties:
            type - network ID
    """

    @property
    @abstractmethod
    def type(self) -> int:
        """
        Get network ID for address

        :return: integer as EntityType
        """
        raise NotImplemented

    @property
    @abstractmethod
    def is_broadcast(self) -> bool:
        # return isinstance(self, BroadcastAddress)
        raise NotImplemented

    @property
    @abstractmethod
    def is_user(self) -> bool:
        # return entity_is_user(network=self.type)
        raise NotImplemented

    @property
    @abstractmethod
    def is_group(self) -> bool:
        # return entity_is_group(network=self.type)
        raise NotImplemented

    #
    #   Factory methods
    #

    @classmethod
    def generate(cls, meta, network: int):  # -> Address:
        gf = general_factory()
        return gf.generate_address(meta=meta, network=network)

    @classmethod
    def create(cls, address: str):  # -> Optional[Address]:
        gf = general_factory()
        return gf.create_address(address=address)

    @classmethod
    def parse(cls, address: Any):  # -> Optional[Address]:
        gf = general_factory()
        return gf.parse_address(address=address)

    @classmethod
    def factory(cls):  # -> Optional[AddressFactory]:
        gf = general_factory()
        return gf.get_address_factory()

    @classmethod
    def register(cls, factory):
        gf = general_factory()
        gf.set_address_factory(factory=factory)


def general_factory():
    from ..factory import AccountFactoryManager
    return AccountFactoryManager.general_factory


class AddressFactory(ABC):

    @abstractmethod
    def generate_address(self, meta, network: int) -> Address:
        """
        Generate address with meta & type

        :param meta: meta info
        :param network: address type
        :return: Address
        """
        raise NotImplemented

    @abstractmethod
    def create_address(self, address: str) -> Optional[Address]:
        """
        Create address from string

        :param address: address string
        :return: Address
        """
        raise NotImplemented

    @abstractmethod
    def parse_address(self, address: str) -> Optional[Address]:
        """
        Parse string object to address

        :param address: address string
        :return: Address
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, List, Any

from ..types import Stringer

from .address import Address


class ID(Stringer, ABC):
    """
        ID for entity (User/Group)
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        data format: "name@address[/terminal]"

        fields:
            name     - entity name, the seed of fingerprint to generate address
            address  - a string to identify an entity
            terminal - entity login resource(device), OPTIONAL
    """

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        raise NotImplemented

    @property
    @abstractmethod
    def address(self) -> Address:
        raise NotImplemented

    @property
    @abstractmethod
    def terminal(self) -> Optional[str]:
        raise NotImplemented

    @property
    @abstractmethod
    def type(self) -> int:
        # return self.address.type
        raise NotImplemented

    @property
    @abstractmethod
    def is_broadcast(self) -> bool:
        # return self.address.is_broadcast
        raise NotImplemented

    @property
    @abstractmethod
    def is_user(self) -> bool:
        # return self.address.is_user
        raise NotImplemented

    @property
    @abstractmethod
    def is_group(self) -> bool:
        # return self.address.is_group
        raise NotImplemented

    @classmethod
    def convert(cls, array: List[str]):  # -> List[ID]:
        """
        Convert ID list from string array

        :param array: string array
        :return: ID list
        """
        gf = general_factory()
        return gf.convert_id_list(array=array)

    @classmethod
    def revert(cls, array) -> List[str]:
        """
        Revert ID list to string array

        :param array: ID list
        :return: string array
        """
        gf = general_factory()
        return gf.revert_id_list(array=array)

    #
    #   Factory methods
    #

    @classmethod
    def generate(cls, meta, network: int, terminal: Optional[str] = None):  # -> ID:
        gf = general_factory()
        return gf.generate_id(meta=meta, network=network, terminal=terminal)

    @classmethod
    def create(cls, name: Optional[str], address: Address, terminal: Optional[str] = None):  # -> ID:
        gf = general_factory()
        return gf.create_id(name=name, address=address, terminal=terminal)

    @classmethod
    def parse(cls, identifier: Any):  # -> Optional[ID]:
        gf = general_factory()
        return gf.parse_id(identifier=identifier)

    @classmethod
    def factory(cls):  # -> Optional[IDFactory]:
        gf = general_factory()
        return gf.get_id_factory()

    @classmethod
    def register(cls, factory):
        gf = general_factory()
        gf.set_id_factory(factory=factory)


def general_factory():
    from ..factory import AccountFactoryManager
    return AccountFactoryManager.general_factory


class IDFactory(ABC):

    @abstractmethod
    def generate_id(self, meta, network: int, terminal: Optional[str]) -> ID:
        """
        Generate ID

        :param meta:     meta info
        :param network:  ID.type
        :param terminal: ID.terminal
        :return: ID
        """
        raise NotImplemented

    @abstractmethod
    def create_id(self, name: Optional[str], address: Address, terminal: Optional[str]) -> ID:
        """
        Create ID

        :param name:     ID.name
        :param address:  ID.address
        :param terminal: ID.terminal
        :return: ID
        """
        raise NotImplemented

    @abstractmethod
    def parse_id(self, identifier: str) -> Optional[ID]:
        """
        Parse string object to ID

        :param identifier: ID string
        :return: ID
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, MutableMapping, Dict, List


class Stringer(ABC):
    """
        Constant String Wrapper
        ~~~~~~~~~~~~~~~~~~~~~~~
    """

    def __hash__(self) -> int:
        return NotImplemented

    def __len__(self) -> int:
        return NotImplemented

    def __eq__(self, other) -> bool:
        return NotImplemented

    def __ne__(self, other) -> bool:
        return NotImplemented

    def __str__(self) -> str:
        return NotImplemented

    @property
    @abstractmethod
    def string(self) -> str:
        """ get inner string """
        return NotImplemented


class Mapper(MutableMapping[str, Any], ABC):
    """
        Mutable Map Wrapper
        ~~~~~~~~~~~~~~~~~~~
    """

    @abstractmethod
    def get_str(self, key: str, default: Optional[str] = None) -> Optional[str]:
        raise NotImplemented

    @abstractmethod
    def get_bool(self, key: str, default: bool = False) -> bool:
        raise NotImplemented

    @abstractmethod
    def get_int(self, key: str, default: int = 0) -> int:
        raise NotImplemented

    @abstractmethod
    def get_float(self, key: str, default: float = 0.0) -> float:
        raise NotImplemented

    @abstractmethod
    def get_time(self, key: str, default: float = 0.0) -> float:
        """ get timestamp in seconds """
        raise NotImplemented

    @abstractmethod
    def set_time(self, key: str, time: Optional[float]):
        """ set timestamp in seconds """
        raise NotImplemented

    @abstractmethod
    def set_string(self, key: str, string: Optional[Stringer]):
        raise NotImplemented

    @abstractmethod
    def set_map(self, key: str, dictionary):  # dictionary: Optional[Mapper]
        raise NotImplemented

    @property
    @abstractmethod
    def dictionary(self) -> Dict[str, Any]:
        """ get inner map """
        raise NotImplemented

    @abstractmethod
    def copy_dictionary(self, deep_copy: bool = False) -> Dict[str, Any]:
        """ copy inner map """
        raise NotImplemented


class Wrapper:

    @classmethod
    def get_str(cls, s) -> Optional[str]:
        """
            Get inner string
            ~~~~~~~~~~~~~~~~
        """
        if isinstance(s, Stringer):
            return s.string
        elif isinstance(s, str):
            return s

    @classmethod
    def get_dict(cls, d) -> Optional[Dict[str, Any]]:
        """
            Get inner map
            ~~~~~~~~~~~~~
            Remove first wrapper
        """
        if isinstance(d, Mapper):
            return d.dictionary
        elif isinstance(d, Dict):
            return d

    @classmethod
    def unwrap(cls, o) -> Any:
        """
            Unwrap object container
            ~~~~~~~~~~~~~~~~~~~~~~~
        """
        if isinstance(o, Stringer):
            return o.string
        elif isinstance(o, Mapper):
            return cls.unwrap_dict(o.dictionary)
        elif isinstance(o, Dict):
            return cls.unwrap_dict(o)
        elif isinstance(o, List):
            return cls.unwrap_list(o)
        return o

    @classmethod
    def unwrap_dict(cls, d) -> Dict[str, Any]:
        """ Unwrap values for keys in map """
        if isinstance(d, Mapper):
            d = d.dictionary
        dictionary = {}
        for key in d:
            value = d[key]
            naked = cls.unwrap(value)
            dictionary[key] = naked
        return dictionary

    @classmethod
    def unwrap_list(cls, a) -> List[Any]:
        """ Unwrap values in the array """
        array = []
        for value in a:
            naked = cls.unwrap(value)
            array.append(naked)
        return array



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union, Any, Dict

from ..types import Mapper
from ..crypto import VerifyKey, SignKey

from .version import MetaType
from .address import Address
from .identifier import ID


class Meta(Mapper, ABC):
    """This class is used to generate entity ID

        User/Group Meta data
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            type: 1,             // meta version
            seed: "moKy",        // user/group name
            key: "{public key}", // PK = secp256k1(SK);
            fingerprint: "..."   // CT = sign(seed, SK);
        }

        algorithm:
            fingerprint = sign(seed, SK);
    """

    @property
    @abstractmethod
    def type(self) -> int:
        """
        Meta algorithm version

            0x01 - username@address
            0x02 - btc_address
            0x03 - username@btc_address
            0x04 - eth_address
            0x05 - username@eth_address
            ...
        """
        raise NotImplemented

    @property
    @abstractmethod
    def key(self) -> VerifyKey:
        """
        Public key (used for signature)

            RSA
            ECC

        :return: public key
        """
        raise NotImplemented

    @property
    @abstractmethod
    def seed(self) -> Optional[str]:
        """
        Seed to generate fingerprint

        :return: ID.name
        """
        raise NotImplemented

    @property
    @abstractmethod
    def fingerprint(self) -> Optional[bytes]:
        """
        Fingerprint to verify ID and public key

            Build: fingerprint = sign(seed, privateKey)
            Check: verify(seed, fingerprint, publicKey)

        :return: signature
        """
        raise NotImplemented

    @abstractmethod
    def generate_address(self, network: int) -> Address:
        """
        Generate Address with network ID

        :param network:  Address.type
        :return: Address
        """
        raise NotImplemented

    @classmethod
    def check(cls, meta) -> bool:
        gf = general_factory()
        return gf.check_meta(meta=meta)

    @classmethod
    def match_id(cls, meta, identifier: ID) -> bool:
        gf = general_factory()
        return gf.meta_match_id(meta=meta, identifier=identifier)

    @classmethod
    def match_key(cls, meta, key: VerifyKey) -> bool:
        gf = general_factory()
        return gf.meta_match_key(meta=meta, key=key)

    #
    #   Factory methods
    #

    @classmethod
    def generate(cls, version: Union[MetaType, int], key: SignKey, seed: Optional[str] = None):  # -> Optional[Meta]:
        gf = general_factory()
        return gf.generate_meta(version=version, key=key, seed=seed)

    @classmethod
    def create(cls, version: Union[MetaType, int], key: VerifyKey,
               seed: Optional[str] = None, fingerprint: Union[bytes, str, None] = None):  # -> Optional[Meta]:
        gf = general_factory()
        return gf.create_meta(version=version, key=key, seed=seed, fingerprint=fingerprint)

    @classmethod
    def parse(cls, meta: Any):  # -> Optional[Meta]:
        gf = general_factory()
        return gf.parse_meta(meta=meta)

    @classmethod
    def factory(cls, version: Union[MetaType, int]):  # -> Optional[MetaFactory]:
        gf = general_factory()
        return gf.get_meta_factory(version=version)

    @classmethod
    def register(cls, version: Union[MetaType, int], factory):
        gf = general_factory()
        gf.set_meta_factory(version=version, factory=factory)


def general_factory():
    from ..factory import AccountFactoryManager
    return AccountFactoryManager.general_factory


class MetaFactory(ABC):

    @abstractmethod
    def generate_meta(self, key: SignKey, seed: Optional[str]) -> Meta:
        """
        Generate meta

        :param key:  private key
        :param seed: ID.name
        :return: Meta
        """
        raise NotImplemented

    @abstractmethod
    def create_meta(self, key: VerifyKey, seed: Optional[str], fingerprint: Union[bytes, str, None]) -> Meta:
        """
        Create meta

        :param key:         public key
        :param seed:        ID.name
        :param fingerprint: sKey.sign(seed)
        :return: Meta
        """
        raise NotImplemented

    @abstractmethod
    def parse_meta(self, meta: Dict[str, Any]) -> Optional[Meta]:
        """
        Parse map object to meta

        :param meta: meta info
        :return: Meta
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional

from ..types import Mapper


class CryptographyKey(Mapper, ABC):
    """Cryptography key with designated algorithm

        Cryptography Key
        ~~~~~~~~~~~~~~~~

        key data format: {
            algorithm : "RSA", // ECC, AES, ...
            data      : "{BASE64_ENCODE}",
            ...
        }
    """

    @property
    @abstractmethod
    def algorithm(self) -> str:
        """
        Get key algorithm name

        :return: algorithm name
        """
        raise NotImplemented

    @property
    @abstractmethod
    def data(self) -> bytes:
        """
        Get key data

        :return: key data
        """
        raise NotImplemented


class EncryptKey(CryptographyKey, ABC):

    @abstractmethod
    def encrypt(self, data: bytes) -> bytes:
        """
        ciphertext = encrypt(plaintext, PW)
        ciphertext = encrypt(plaintext, PK)

        :param data: plaintext
        :return:     ciphertext
        """
        raise NotImplemented


class DecryptKey(CryptographyKey, ABC):

    @abstractmethod
    def decrypt(self, data: bytes) -> Optional[bytes]:
        """
        plaintext = decrypt(ciphertext, PW);
        plaintext = decrypt(ciphertext, SK);

        :param data: ciphertext
        :return:     plaintext
        """
        raise NotImplemented

    @abstractmethod
    def match(self, key: EncryptKey) -> bool:
        """
        OK = decrypt(encrypt(data, SK), PK) == data

        :param key: encrypt key
        :return:    False on error
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

from .asymmetric import SignKey
from .public import PublicKey


class PrivateKey(SignKey, ABC):
    """This class is used to decrypt symmetric key or sign message data

        Asymmetric Cryptography Private Key
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        key data format: {
            algorithm : "RSA", // "ECC", ...
            data      : "{BASE64_ENCODE}",
            ...
        }
    """

    @property
    @abstractmethod
    def public_key(self) -> PublicKey:
        """
        Get public key from private key

        :return: public key paired to this private key
        """
        raise NotImplemented

    #
    #  Factory methods
    #

    @classmethod
    def generate(cls, algorithm: str):  # -> Optional[PrivateKey]:
        gf = general_factory()
        return gf.generate_private_key(algorithm=algorithm)

    @classmethod
    def parse(cls, key: Any):  # -> Optional[PrivateKey]:
        gf = general_factory()
        return gf.parse_private_key(key=key)

    @classmethod
    def factory(cls, algorithm: str):  # -> Optional[PrivateKeyFactory]:
        gf = general_factory()
        return gf.get_private_key_factory(algorithm=algorithm)

    @classmethod
    def register(cls, algorithm: str, factory):
        gf = general_factory()
        gf.set_private_key_factory(algorithm=algorithm, factory=factory)


def general_factory():
    from .factory import CryptographyKeyFactoryManager
    return CryptographyKeyFactoryManager.general_factory


class PrivateKeyFactory(ABC):

    @abstractmethod
    def generate_private_key(self) -> PrivateKey:
        """
        Generate key

        :return: PrivateKey
        """
        raise NotImplemented

    @abstractmethod
    def parse_private_key(self, key: Dict[str, Any]) -> Optional[PrivateKey]:
        """
        Parse map object to key

        :param key: key info
        :return: PrivateKey
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from .types import ConstantString
from .protocol import ID, Address

from .address import ANYWHERE, EVERYWHERE


class Identifier(ConstantString, ID):

    def __init__(self, identifier: str, name: Optional[str], address: Address, terminal: Optional[str] = None):
        super().__init__(string=identifier)
        self.__name = name
        self.__address = address
        self.__terminal = terminal

    @property  # Override
    def name(self) -> Optional[str]:
        return self.__name

    @property  # Override
    def address(self) -> Address:
        return self.__address

    @property  # Override
    def terminal(self) -> Optional[str]:
        return self.__terminal

    @property  # Override
    def type(self) -> int:
        return self.address.type

    @property  # Override
    def is_broadcast(self) -> bool:
        return self.address.is_broadcast

    @property  # Override
    def is_user(self) -> bool:
        return self.address.is_user

    @property  # Override
    def is_group(self) -> bool:
        return self.address.is_group


"""
    ID for Broadcast
    ~~~~~~~~~~~~~~~~
"""

ANYONE = Identifier(identifier='anyone@anywhere', name='anyone', address=ANYWHERE)
EVERYONE = Identifier(identifier='everyone@everywhere', name='everyone', address=EVERYWHERE)

# DIM Founder
FOUNDER = Identifier(identifier='moky@anywhere', name='moky', address=ANYWHERE)



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import copy
from typing import Any, Optional, Iterator, Tuple, Dict
from typing import Mapping, ItemsView, KeysView, ValuesView

from mkm.types import Stringer
from .converter import Converter
from .wrapper import Mapper


class Dictionary(Mapper):
    """
        Mutable Map Wrapper
        ~~~~~~~~~~~~~~~~~~~
        A container sharing the same inner dictionary
    """

    def __init__(self, dictionary: Dict = None):
        super().__init__()
        if dictionary is None:
            dictionary = {}
        elif isinstance(dictionary, Mapper):
            dictionary = dictionary.dictionary
        self.__dictionary = dictionary

    @property  # Override
    def dictionary(self) -> Dict:
        return self.__dictionary

    # Override
    def copy_dictionary(self, deep_copy: bool = False) -> Dict:
        if deep_copy:
            copy.deepcopy(self.__dictionary)
        else:
            return self.__dictionary.copy()

    # Override
    def clear(self):
        """ D.clear() -> None.  Remove all items from D. """
        self.__dictionary.clear()

    # Override
    def copy(self):
        """ D.copy() -> a shallow copy of D """
        dictionary = self.__dictionary.copy()
        return Dictionary(dictionary=dictionary)

    @staticmethod
    def fromkeys(*args, **kwargs):
        """ Create a new dictionary with keys from iterable and values set to value. """
        dictionary = dict.fromkeys(*args, **kwargs)
        return Dictionary(dictionary=dictionary)

    # Override
    def get_str(self, key: str, default: Optional[str] = None) -> Optional[str]:
        value = self.__dictionary.get(key, default)
        value = Converter.get_str(value=value)
        return default if value is None else value

    # Override
    def get_bool(self, key: str, default: bool = False) -> bool:
        value = self.__dictionary.get(key, default)
        value = Converter.get_bool(value=value)
        return default if value is None else value

    # Override
    def get_int(self, key: str, default: int = 0) -> int:
        value = self.__dictionary.get(key, default)
        value = Converter.get_int(value=value)
        return default if value is None else value

    # Override
    def get_float(self, key: str, default: float = 0.0) -> float:
        value = self.__dictionary.get(key, default)
        value = Converter.get_float(value=value)
        return default if value is None else value

    # Override
    def get_time(self, key: str, default: float = 0.0) -> float:
        value = self.__dictionary.get(key, default)
        value = Converter.get_time(value=value)
        return default if value is None else value

    # Override
    def set_time(self, key: str, time: Optional[float]):
        if time is None:
            self.__dictionary.pop(key, None)
        else:
            self.__dictionary[key] = time

    # Override
    def set_string(self, key: str, string: Optional[Stringer]):
        if string is None:
            self.__dictionary.pop(key, None)
        else:
            self.__dictionary[key] = string.string

    # Override
    def set_map(self, key: str, dictionary: Optional[Mapper]):
        if dictionary is None:
            self.__dictionary.pop(key, None)
        else:
            self.__dictionary[key] = dictionary.dictionary

    # Override
    def get(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """ Return the value for key if key is in the dictionary, else default. """
        return self.__dictionary.get(key, default)

    # Override
    def items(self) -> ItemsView[str, Any]:
        """ D.items() -> a set-like object providing a view on D's items """
        return self.__dictionary.items()

    # Override
    def keys(self) -> KeysView[str]:
        """ D.keys() -> a set-like object providing a view on D's keys """
        return self.__dictionary.keys()

    # Override
    def pop(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        return self.__dictionary.pop(key, default)

    # Override
    def popitem(self) -> Tuple[str, Any]:
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if D is empty.
        """
        return self.__dictionary.popitem()

    # Override
    def setdefault(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Insert key with a value of default if key is not in the dictionary.

        Return the value for key if key is in the dictionary, else default.
        """
        self.__dictionary.setdefault(key, default)

    # Override
    def update(self, __m: Mapping[str, Any], **kwargs: Any):
        self.__dictionary.update(__m, **kwargs)

    # Override
    def values(self) -> ValuesView[Any]:
        """ D.values() -> an object providing a view on D's values """
        return self.__dictionary.values()

    def __contains__(self, o) -> bool:
        """ True if the dictionary has the specified key, else False. """
        return self.__dictionary.__contains__(o)

    def __delitem__(self, v: str):
        """ Delete self[key]. """
        self.__dictionary.__delitem__(v)

    def __eq__(self, o: Dict) -> bool:
        """ Return self==value. """
        if isinstance(o, Mapper):
            if self is o:
                # same object
                return True
            o = o.dictionary
        # check inner map
        return self.__dictionary.__eq__(o)

    # def __getattribute__(self, name: str) -> Any:
    #     """ Return getattr(self, name). """
    #     if isinstance(name, Stringer):
    #         name = name.string
    #     return self.__dictionary.__getattribute__(name=name)

    def __getitem__(self, k: str) -> Any:
        """ x.__getitem__(y) <==> x[y] """
        return self.__dictionary.__getitem__(k)

    def __ge__(self, other) -> bool:
        """ Return self>=value. """
        pass

    def __gt__(self, other) -> bool:
        """ Return self>value. """
        pass

    def __iter__(self) -> Iterator[str]:
        """ Implement iter(self). """
        return self.__dictionary.__iter__()

    def __len__(self) -> int:
        """ Return len(self). """
        return self.__dictionary.__len__()

    def __le__(self, other) -> bool:
        """ Return self<=value. """
        pass

    def __lt__(self, other) -> bool:
        """ Return self<value. """
        pass

    def __ne__(self, o: Dict) -> bool:
        """ Return self!=value. """
        if isinstance(o, Mapper):
            if self is o:
                # same object
                return False
            o = o.dictionary
        # check inner map
        return self.__dictionary.__ne__(o)

    def __repr__(self) -> str:
        """ Return repr(self). """
        return self.__dictionary.__repr__()

    def __setitem__(self, k: str, v: Optional[Any]):
        """ Set self[key] to value. """
        self.__dictionary.__setitem__(k, v)

    def __sizeof__(self) -> int:
        """ D.__sizeof__() -> size of D in memory, in bytes """
        return self.__dictionary.__sizeof__()

    __hash__ = None



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC
from typing import Any, Optional


class Converter(ABC):

    @classmethod
    def get_str(cls, value: Any) -> Optional[str]:
        if value is None:
            return None
        elif isinstance(value, str):
            # exactly
            return value
        else:
            return str(value)

    @classmethod
    def get_bool(cls, value: Any) -> Optional[bool]:
        """ assume value can be a config string:
            'true', 'false', 'yes', 'no', 'on', 'off', '1', '0', ...
        """
        if value is None:
            return None
        elif isinstance(value, bool):
            # exactly
            return value
        elif isinstance(value, int):
            return value != 0
        elif isinstance(value, float):
            return value != 0.0
        elif isinstance(value, str):
            lo = value.lower()
            # return lo in true_array
            return lo not in false_array
        else:
            return True

    @classmethod
    def get_int(cls, value: Any) -> Optional[int]:
        if value is None:
            return None
        elif isinstance(value, int):
            # exactly
            return value
        elif isinstance(value, bool) or isinstance(value, float) or isinstance(value, str):
            return int(value)
        else:
            return 0

    @classmethod
    def get_float(cls, value: Any) -> Optional[float]:
        if value is None:
            return None
        elif isinstance(value, float):
            # exactly
            return value
        elif isinstance(value, bool) or isinstance(value, int) or isinstance(value, str):
            return float(value)
        else:
            return 0.0

    @classmethod
    def get_time(cls, value: Any) -> Optional[float]:
        """ assume value be a timestamp in seconds
            (from 1970-01-01 00:00:00)
        """
        return cls.get_float(value=value)


true_array = [
    'true', 'yes', 'on', '1',
]
false_array = [
    'false', 'no', 'off', '0', 'null', 'undefined',
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Ming-Ke-Ming
    ~~~~~~~~~~~~

    Common Identity Module for decentralized user identity authentication
"""

import io

from setuptools import setup, find_packages

__version__ = '0.12.8'
__author__ = 'Albert Moky'
__contact__ = 'albert.moky@gmail.com'

with io.open('README.md', 'r', encoding='utf-8') as fh:
    readme = fh.read()

setup(
    name='mkm',
    version=__version__,
    url='https://github.com/dimchat/mkm-py',
    license='MIT',
    author=__author__,
    author_email=__contact__,
    description='A common identity module',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
    ]
)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

from ..crypto import VerifyKey, SignKey


class TAI(ABC):
    """
        The Additional Information
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        'Meta' is the information for entity which never changed,
            which contains the key for verify signature;
        'TAI' is the variable part,
            which could contain a public key for asymmetric encryption.
    """

    @property
    @abstractmethod
    def valid(self) -> bool:
        """
        Check if signature matched

        :return: True on matched
        """
        raise NotImplemented

    #
    #  signature
    #

    @abstractmethod
    def verify(self, public_key: VerifyKey) -> bool:
        """
        Verify 'data' and 'signature' with public key

        :param public_key: public key in meta.key
        :return: True on signature matched
        """
        raise NotImplemented

    @abstractmethod
    def sign(self, private_key: SignKey) -> Optional[bytes]:
        """
        Encode properties to 'data' and sign it to 'signature'

        :param private_key: private key match meta.key
        :return: signature, None on error
        """
        raise NotImplemented

    #
    #  properties
    #

    @property
    @abstractmethod
    def properties(self) -> Optional[Dict[str, Any]]:
        """
        Get all properties when valid

        :return: inner dictionary
        """
        raise NotImplemented

    @abstractmethod
    def get_property(self, key: str) -> Optional[Any]:
        """
        Get property value with key

        :param key: property key
        :return: property value
        """
        raise NotImplemented

    @abstractmethod
    def set_property(self, key: str, value: Optional[Any]):
        """
        Update property with key and data
        (this will clear 'data' and 'signature')

        :param key:   property key
        :param value: property value
        """
        raise NotImplemented
